from django.db import models
from django.utils.translation import ugettext_lazy as _


class FileMetadata(models.Model):
    inode = models.IntegerField(db_index=True, verbose_name=_('inode'))
    filename = models.TextField(blank=True, null=True, verbose_name=_('Filename'))
    suffix = models.CharField(blank=True, null=True, max_length=10, verbose_name=_('Suffix'))
    path = models.TextField(blank=True, null=True, verbose_name=_('Path'))
    size = models.IntegerField(blank=True, null=True, verbose_name=_('Size (bits)'))
    size_str = models.CharField(blank=True, null=True, max_length=10, verbose_name=_('Size'))
    created = models.DateTimeField(blank=True, auto_now_add=True, verbose_name=_('Created'))
    modified = models.DateTimeField(blank=True, auto_now_add=True, verbose_name=_('Modified'))
    title = models.CharField(blank=True, null=True, max_length=256, verbose_name=_('Title'))
    description = models.CharField(blank=True, null=True, max_length=256, verbose_name=_('Description'))
    text_content = models.TextField(blank=True, null=True, verbose_name=_('Text content'))

    class Meta:
        verbose_name = _('static file')
        verbose_name_plural = _('static files')

    def __repr__(self):
        # return '<StaticPage: %s -- %s>' % (self.url, truncatewords_html(self.content, 10))
        return self.name



import datetime
from pathlib import Path
import os
from django.conf import settings

from filemetadata.models import FileMetadata
from filemetadata.utils import human_readable_size
from filemetadata.utils import is_binary_string
from tzlocal import get_localzone
import logging

logger = logging.getLogger("django-filemetadata")

try:
    import PyPDF4

    PYPDF4_LOADED = True
except ImportError:
    PYPDF4_LOADED = False
    logger.warn("PyPDF4 library not installed. It'll not possible to extract the text from '.pdf' files")


def func_extract_text(file_obj):
    out = ""
    with file_obj.open(mode='rb') as file_data:
        if file_obj.suffix == '.pdf':
            if PYPDF4_LOADED:
                lines = []
                pdf = PyPDF4.PdfFileReader(file_data)
                for num_page in range(0, pdf.getNumPages()):
                    lines.append(pdf.getPage(num_page).extractText())
                out = "\n".join(lines)
        else:
            if not is_binary_string(file_data.read(1024)):
                # Try to index all file in txt format.
                file_data.seek(0)
                out = file_data.read()
    return out


class FileIndexer(object):

    def extract_text(self, file_obj):
        """
        Extract the text from the file object
        :param file_obj: file object
        :return: text content
        """
        return self.sanitize_text(func_extract_text(file_obj))

    @staticmethod
    def sanitize_text(txt):
        """
        Converto to utf-8 and remove unnecessary data and empty lines.
        :param txt: text
        :return: compact text (utf-8)
        """
        txtout = []
        if type(txt) == bytes:
            txt = txt.decode("utf-8", "replace")
        for line in txt.replace(u"\xa0", " ").split("\n"):
            line = line.strip()
            if line:
                txtout.append(line)
        return "\n".join(txtout)

    def normalize_dirs(self, dirs):
        if not dirs:
            dirs = settings.FILEMETADATA_LOOKUP_DIRS
        if not dirs:
            raise FileNotFoundError("Please informe the folder(s) to index or "
                                    "configure FILEMETADATA_LOOKUP_DIRS in the settings.")

        ret = []
        if not isinstance(dirs, (list, tuple)):
            dirs = dirs.split(";")
        for d in dirs:
            error = ''
            d = os.path.abspath(d.strip())
            if os.path.exists(d):
                if os.path.isdir(d):
                    ret.append(d)
                else:
                    error = "'%s' is not a directory." % d
            else:
                error = "'%s' does not exists." % d

            if error:
                if self.raise_errors:
                    raise FileNotFoundError(error)
                else:
                    self.errors.append(error)
        return ret

    def __init__(self, follow_symlinks=True, extract_content=False, recursive=True, raise_errors=False,
                 verbose_print=False):
        """
        Look for files in the filessystem, extract the metadata and save into the DB.

        :param follow_symlinks: follow the symlinks
        :param extract_content: try to get the content of the file in text format.
        :param recursive: search the file recursively in the directories
        :param recursive:
        """
        self.follow_symlinks = follow_symlinks
        self.extract_content = extract_content
        self.recursive = recursive
        self.raise_errors = raise_errors
        self.verbose_print = verbose_print
        self.errors = []
        self.count_added = 0
        self.count_updated = 0
        self.count_deleted = 0

    def retrieve_metadata_file(self, file_obj):
        """
        This method return the metadata from the file.

        :param file_obj: The file (PosixPath object)
        :return: dictionary with the metadata
        """

        filename = file_obj.name
        # TODO: Add follow_symlinks?
        stat = file_obj.stat()
        fpath = os.path.abspath(os.path.dirname(file_obj))
        local = get_localzone()

        content_txt = ""
        if self.extract_content:
            try:
                content_txt = self.extract_text(file_obj)
            except KeyboardInterrupt:
                raise
            except Exception as e:
                if self.raise_errors:
                    raise
                else:
                    self.errors.append("Error extracting the text from file '%s'. %s" % (filename, e))

        return {'file': file_obj,
                'suffix': file_obj.suffix,
                'inode': stat.st_ino,
                'filename': filename,
                'path': fpath,
                'size': stat.st_size,
                'size_str': human_readable_size(stat.st_size),
                'created': datetime.datetime.fromtimestamp(stat.st_ctime, tz=local),
                'modified': datetime.datetime.fromtimestamp(stat.st_mtime, tz=local),
                'title': filename,
                'text_content': content_txt
                }

    def retrieve_metadata_files(self, dirs_str):
        """
        This method return all metadata from the files found in the directories.

        :param dirs_str: list of directories. eg. ['/var/folder1','/var/folder2',]
        :return: list (dictionary) with all metadata.
        """

        def print_indented(obj, prefix, level):
            if self.verbose_print:
                print('  ' * level + prefix + obj)

        def retrieve_folder(dir_obj, level=0, rec=True):
            print_indented(dir_obj.name, '= ', level)
            output = []

            if dir_obj.exists():
                for element in dir_obj.iterdir():
                    if element.is_dir():
                        if rec:
                            output += retrieve_folder(element, level=level + 1)
                    else:
                        try:
                            output.append(self.retrieve_metadata_file(element))
                        except KeyboardInterrupt:
                            raise
                        except Exception as e:
                            if self.raise_errors:
                                raise
                            else:
                                self.errors.append("Error extracting metadata from file %s - %s" % (element.name, e))

                        print_indented(element.name, '- ', level + 1)
            else:
                if self.raise_errors:
                    raise FileNotFoundError("'%s' not found" % dir_obj.name)
                else:
                    self.errors.append("'%s' not found" % dir_obj.name)
            return output

        files = []
        dirs_str = self.normalize_dirs(dirs_str)
        for dir_str in dirs_str:
            directory = Path(dir_str.strip())
            files += retrieve_folder(directory, rec=self.recursive)
        return files

    def delete_metadata_db(self, dirs_str=None):
        """
        Delete the data from the DB with the specific folders (path)
        :param dirs_str: directories to delete
        :return: Number of deleted records
        """
        dirs_str = self.normalize_dirs(dirs_str)
        c = 0
        for dir_str in dirs_str:
            dir_str = dir_str.strip()

            if self.recursive:
                pars = {'path__startswith': dir_str}
            else:
                pars = {'path': dir_str}

            for o in FileMetadata.objects.filter(**pars):
                o.delete()
                c += 1
        return c

    def index_metadata_db(self, dirs_str=None, reindex=False):
        """
        Get the metadata from all directories (dirs_str) and save in the DB.

        :param reindex: delete all data and reindex
        :param dirs_str: list of directories. eg. ['/var/folder1','/var/folder2',]
        :return:
        """

        self.errors = []
        dirs_str = self.normalize_dirs(dirs_str)
        self.count_added = 0
        self.count_updated = 0
        self.count_deleted = 0

        if reindex:
            # FileMetadata.objects.all().delete;
            self.count_deleted = self.delete_metadata_db(dirs_str)

        for m in self.retrieve_metadata_files(dirs_str=dirs_str):
            try:
                if m:
                    try:
                        obj = FileMetadata.objects.get(path=m['path'], filename=m['filename'])
                        self.count_updated += 1
                    except FileMetadata.DoesNotExist:
                        self.count_added += 1
                        obj = FileMetadata()

                    # Set the field values
                    fields = [f.name for f in FileMetadata._meta.get_fields()]
                    for k, v in m.items():
                        if k in fields:
                            setattr(obj, k, v)
                    obj.save()
            except KeyboardInterrupt:
                raise
            except Exception as e:
                if self.raise_errors:
                    raise
                else:
                    self.errors.append("Error saving file %s - %s" % (m.get('filename', '[None]'), e))



# This file sets up and configures Django. It's used by scripts that need to
# execute as if running in a Django server.
import os
import django
from django.conf import settings

BASE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "filemetadata"))

def boot_django():
    settings.configure(
        BASE_DIR=BASE_DIR,
        DEBUG=True,
        DATABASES={
            "default":{
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
            }
        },
        INSTALLED_APPS=(
            "filemetadata",
        ),
        TIME_ZONE="UTC",
        USE_TZ=True,
        FILEMETADATA_LOOKUP_DIRS=os.path.join(BASE_DIR, "filemetadata", "tests"),
    )
    django.setup()


from django.contrib import admin

from .models import FileMetadata

@admin.register(FileMetadata)
class FileMetadataAdmin(admin.ModelAdmin):
    list_display = ('filename', 'path', 'size_str')
    list_filter = ('path', 'suffix')




import os

from django.test import TestCase

from filemetadata.indexer import FileIndexer
from filemetadata.models import FileMetadata


def sort(data):
    data.sort()
    return data


class StaticFileTest(TestCase):

    def test_retrieve_metadata_files(self):
        test_folder_a = os.path.join("filemetadata", "tests", "folder_A")
        test_folder_c = os.path.join("filemetadata", "tests", "folder_C")

        file_indexer = FileIndexer(extract_content=True, recursive=True, raise_errors=False)
        res = file_indexer.retrieve_metadata_files(dirs_str=test_folder_a)
        self.assertEqual(len(file_indexer.errors), 0)
        self.assertEqual(len(res), 3)

        # Prepare the results for tests
        fields2test = ['filename', 'size', 'size_str', 'title', 'text_content']
        res_test = {}
        for r in res:
            res_test[r['filename']] = r

        self.assertEqual(
            [(k, v) for k, v in res_test['test_file3.txt'].items() if k in fields2test],
            [('filename', 'test_file3.txt'),
             ('size', 30),
             ('size_str', '30.00 bytes'),
             ('title', 'test_file3.txt'),
             ('text_content', b"This is\nMy Test 3.\nThat's all!")]
        )

        self.assertEqual(
            [(k, v) for k, v in res_test['test_file1.txt'].items() if k in fields2test],
            [('filename', 'test_file1.txt'),
             ('size', 26),
             ('size_str', '26.00 bytes'),
             ('title', 'test_file1.txt'),
             ('text_content', b'This is\nMy Test 1.\nThanks!')]

        )

        self.assertEqual(
            [(k, v) for k, v in res_test['test_file2.txt'].items() if k in fields2test],
            [('filename', 'test_file2.txt'),
             ('size', 27),
             ('size_str', '27.00 bytes'),
             ('title', 'test_file2.txt'),
             ('text_content', b'This is\nMy Test 2.\nSee You!')]
        )

        file_indexer = FileIndexer(extract_content=False)
        res = file_indexer.retrieve_metadata_files(dirs_str=test_folder_c)
        self.assertEqual(
            [(k, v) for k, v in res[0].items() if k in ['filename', 'size', 'size_str', 'title', 'text_content']],
            [('filename', 'test_file4.txt'),
             ('size', 24),
             ('size_str', '24.00 bytes'),
             ('title', 'test_file4.txt'),
             ('text_content', '')]
        )

    def test_index_metadata_db(self):
        """
        Testing sync a file.
        :return:
        """

        test_folder_a = os.path.join("filemetadata", "tests", "folder_A")
        test_folder_c = os.path.join("filemetadata", "tests", "folder_C")

        file_indexer = FileIndexer(extract_content=True, recursive=True, raise_errors=False)
        file_indexer.index_metadata_db(dirs_str=test_folder_a, reindex=False)
        self.assertEqual(file_indexer.count_added, 3)
        self.assertEqual(file_indexer.count_deleted, 0)
        self.assertEqual(file_indexer.count_updated, 0)
        self.assertEqual(len(file_indexer.errors), 0)
        self.assertEqual(sort([o.filename for o in FileMetadata.objects.all()]),
                         ['test_file1.txt', 'test_file2.txt', 'test_file3.txt'])

        # Test reindex
        file_indexer.index_metadata_db(dirs_str=test_folder_a, reindex=True)
        self.assertEqual(file_indexer.count_added, 3)
        self.assertEqual(file_indexer.count_deleted, 3)
        self.assertEqual(file_indexer.count_updated, 0)
        self.assertEqual(len(file_indexer.errors), 0)
        self.assertEqual(sort([o.filename for o in FileMetadata.objects.all()]),
                         ['test_file1.txt', 'test_file2.txt', 'test_file3.txt'])

        # Test multiple folders
        file_indexer.index_metadata_db(dirs_str=[test_folder_a, test_folder_c], reindex=False)
        self.assertEqual(file_indexer.count_added, 1)
        self.assertEqual(file_indexer.count_deleted, 0)
        self.assertEqual(file_indexer.count_updated, 3)
        self.assertEqual(len(file_indexer.errors), 0)
        self.assertEqual(sort([o.filename for o in FileMetadata.objects.all()]),
                         ['test_file1.txt', 'test_file2.txt', 'test_file3.txt', 'test_file4.txt'])

        # Test delete (non recursive)
        file_indexer = FileIndexer(recursive=False)
        deleted = file_indexer.delete_metadata_db(dirs_str=test_folder_a)
        self.assertEqual(deleted, 2)
        self.assertEqual(sort([o.filename for o in FileMetadata.objects.all()]),
                         ['test_file3.txt', 'test_file4.txt'])

        # Test delete (recursive)
        file_indexer = FileIndexer(recursive=True)
        deleted = file_indexer.delete_metadata_db([test_folder_a, test_folder_c])
        self.assertEqual(deleted, 2)
        self.assertEqual(sort([o.filename for o in FileMetadata.objects.all()]),
                         [])

    def test_index_metadata_db_options(self):
        test_folder_a = os.path.join("filemetadata", "tests", "folder_A")
        test_file1 = os.path.join("filemetadata", "tests", "folder_A", "test_file1.txt")

        # Test non-recursive
        file_indexer = FileIndexer(recursive=False)
        file_indexer.index_metadata_db(dirs_str=test_folder_a)
        self.assertEqual(file_indexer.count_added, 2)
        self.assertEqual(sort([o.filename for o in FileMetadata.objects.all()]),
                         ['test_file1.txt', 'test_file2.txt'])

        # Test raise_errors
        file_indexer.raise_errors = True
        self.assertRaises(FileNotFoundError, file_indexer.index_metadata_db, dirs_str=[test_file1,
                                                                                       test_folder_a,
                                                                                       'wrong_folder_error'])

        # Test raise_errors
        file_indexer.raise_errors = False
        file_indexer.index_metadata_db(dirs_str=[test_file1,
                                                 test_folder_a,
                                                 'wrong_folder_error'])
        self.assertEqual(len(file_indexer.errors), 2)



import sys
from unittest import TestSuite
from boot_django import boot_django

boot_django()

default_labels = ["filemetadata.tests", ]

def get_suite(labels=default_labels):
    from django.test.runner import DiscoverRunner
    runner = DiscoverRunner(verbosity=1)
    failures = runner.run_tests(labels)
    if failures:
        sys.exit(failures)

    # In case this is called from setuptools, return a test suite
    return TestSuite()

if __name__ == "__main__":
    labels = default_labels
    if len(sys.argv[1:]) > 0:
        labels = sys.argv[1:]

    get_suite(labels)


import os
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name="django-filemetadata",
    version="1.0.4",
    author="Rodrigo Ristow",
    author_email="rodrigo@maxttor.com",
    description="Synchronize the metadata from local files in the DB",
    license="BSD",
    keywords="django file metadata",
    url="https://gitlab.com/rristow/django-filemetadata",
    packages=find_packages(),
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
    ],
)



# Generated by Django 2.2.13 on 2020-12-30 18:47

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FileMetadata',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('inode', models.IntegerField(db_index=True, verbose_name='inode')),
                ('filename', models.TextField(blank=True, null=True, verbose_name='Filename')),
                ('suffix', models.CharField(blank=True, max_length=10, null=True, verbose_name='Suffix')),
                ('path', models.TextField(blank=True, null=True, verbose_name='Path')),
                ('size', models.IntegerField(blank=True, null=True, verbose_name='Size (bits)')),
                ('size_str', models.CharField(blank=True, max_length=10, null=True, verbose_name='Size')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='Created')),
                ('modified', models.DateTimeField(auto_now_add=True, verbose_name='Modified')),
                ('title', models.CharField(blank=True, max_length=256, null=True, verbose_name='Title')),
                ('description', models.CharField(blank=True, max_length=256, null=True, verbose_name='Description')),
                ('text_content', models.TextField(blank=True, null=True, verbose_name='Text content')),
            ],
            options={
                'verbose_name': 'static file',
                'verbose_name_plural': 'static files',
            },
        ),
    ]





def is_binary_string(bytes_str):
    """ Use a heuristic to check if the bytes_str are binary data (not text).
    Attention: It will not works in all cases.
    """
    textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})
    return bool(bytes_str.translate(None, textchars))


def human_readable_size(size, decimal_places=2):
    """
    Convert the integer bits in a human readable string with Mb, Gb and so on.
    :param size: bits (integer)
    :param decimal_places: decimals to show
    :return: Size in string format.
    """
    for unit in ['bytes', 'Kb', 'Mb', 'Gb', 'Tb', 'Pb']:
        if size < 1024.0 or unit == 'Pb':
            break
        size /= 1024.0
    return f"{size:.{decimal_places}f} {unit}"



from django.core.management.base import BaseCommand

from filemetadata.indexer import FileIndexer


class Command(BaseCommand):
    help = 'Update the the file-metadata found in the directories into the DB.'

    def add_arguments(self, parser):
        parser.add_argument('-f', dest='folders', type=str, required=False,
                            help="Folder(s) to index (coma separated)")
        parser.add_argument('-c', dest='clear', default=False, action='store_true',
                            help="Clear the data before reindex")
        parser.add_argument('-d', dest='delete', default=False, action='store_true',
                            help="Delete only the data from these folders and exit")
        parser.add_argument('-s', dest='symlinks', default=False, action='store_true',
                            help="Index the symlinks (Do not follow it)")
        parser.add_argument('-x', dest='content', default=False, action='store_true',
                            help="Extract the content of the file (text)")
        parser.add_argument('-n', dest='non_reentrant', default=False, action='store_true',
                            help="Non-reentrant mode (Not recursive)")
        parser.add_argument('-a', dest='abort', default=False, action='store_true',
                            help="Abort on errors")

    def handle(self, *args, **options):
        folders_str = None
        if options.get('folders', None):
            # TODO: find a better solution to allow ',' in the name of the folder
            folders_str = options['folders'].split(",")
        file_indexer = FileIndexer(follow_symlinks=not options.get('symlinks', False),
                                   extract_content=options.get('content', False),
                                   recursive=not options.get('non_reentrant', False),
                                   raise_errors=options.get('abort', False),
                                   verbose_print=True)
        if options.get('delete', False):
            file_indexer.count_deleted = file_indexer.delete_metadata_db(dirs_str=folders_str)
        else:
            file_indexer.index_metadata_db(dirs_str=folders_str,
                                           reindex=options.get('clear', False))

        if file_indexer.errors:
            msg = "Process executed with %s errors!" % len(file_indexer.errors)
        else:
            msg = "Process successful!"

        print("%s (deleted: %s, added: %s, updated: %s)\n%s" % (
            msg,
            file_indexer.count_deleted,
            file_indexer.count_added,
            file_indexer.count_updated,
            '\n'.join(file_indexer.errors)))
